----------------------------------
-- DISTRIBUTOR
----------------------------------
--[[
Basic Idea:

Sender:
	Update available (->Raid)
	Collect Responses (<- Whisper)
	Start Upload (->Whisper)
	Collect Success (<-Whisper)

Receiver:
	Listen for Update (<-Raid)
	Respond (->Whisper), Listen for Upload
	Download (<-Whisper)
	Success (->Whisper)

]]

-- Credits to Bazaar (by Shadowed) for this idea

local defaults = {
	profile = {
		AutoAccept = true,
	},
}

local addon = DXE
local L = addon.L

local Colors = addon.Media.Colors

local ipairs, pairs = ipairs, pairs
local remove,wipe = table.remove,table.wipe
local match,len,format,split,find = string.match,string.len,string.format,string.split,string.find

local db,pfl

----------------------------------
-- CONSTANTS
----------------------------------

local MAIN_PREFIX = "DXE_D"
local TRANSFER_PREFIX = "DXE_T"
local UL_WAIT = 5
local DL_TIMEOUT = 35

----------------------------------
-- INITIALIZATION
----------------------------------

local module = addon:NewModule("Distributor","AceEvent-3.0","AceTimer-3.0","AceComm-3.0","AceSerializer-3.0")
addon.Distributor = module

function module:RefreshProfile() pfl = self.db.profile end

function module:OnInitialize()
	StackAnchor = addon:CreateLockableFrame("DistributorStackAnchor",200,10,L["Download/Upload Anchor"])
	addon:RegisterMoveSaving(StackAnchor,"CENTER","UIParent","CENTER",0,300)
	addon:LoadPosition("DXEDistributorStackAnchor")

	self.db = addon.db:RegisterNamespace("Distributor", defaults)
	db = self.db
	pfl = db.profile

	db.RegisterCallback(self, "OnProfileChanged", "RefreshProfile")
	db.RegisterCallback(self, "OnProfileCopied", "RefreshProfile")
	db.RegisterCallback(self, "OnProfileReset", "RefreshProfile")
end

function module:OnEnable()
	self:RegisterComm(MAIN_PREFIX)
end

----------------------------------
-- INITIAL DISTRIBUTE
----------------------------------
-- The active downloads
local Downloads = {}

-- The active uploads
local Uploads = {}
-- The queued uploads
local UploadQueue = {}


-- Used to start uploads
function module:Distribute(key)
	wipe(Uploads)
	wipe(UploadQueue)
	local channel = "RAID" -- I don't think we need any other
	local data = addon.EDB[key]

	if not data or key == "default" or GetNumRaidMembers() == 0 then
		return
	end

	local serialData = self:Serialize(data)
	local length = len(serialData)

	local message = format("UPDATE:%s:%d:%d:%s",key,data.version,length,data.name) -- ex. UPDATE:sartharion:150:800:Sartharion

	Uploads[key] = {
		accepts = 0,
		declines = 0,
		sent = 0,
		serialData = serialData,
		length = length,
		name = data.name,
	}

	self:SendCommMessage(MAIN_PREFIX, message, channel)
end


----------------------------------
-- RESPONDER
----------------------------------
function module:Respond(msg,sender)
	self:SendCommMessage(MAIN_PREFIX, msg, "WHISPER",sender)
end

----------------------------------
-- MAIN
----------------------------------
function module:Main(msg, dist, sender)
	local type,args = match(msg,"(%w+):(.+)")

	-- Someone wants to send an encounter
	if type == "UPDATE" then
		local key,version,length,name = split(":",args)
		version = tonumber(version)

		local data = addon.EDB[key]
		-- Version check
		if (data and data.version >= version) then
			self:Respond(format("RESPONSE:%s:%s",key,"NO"),sender)
			return
		end

		if pfl.AutoAccept then
			self:StartReceiving(key,sender,length,name)
			self:Respond(format("RESPONSE:%s:%s",key,"YES"),sender)
			return
		end

		local popupkey = format("DXE_Confirm_%s",key)
		if not StaticPopupDialogs[popupkey] then
			local STATIC_CONFIRM = {
				text = format(L["%s is sharing an update for %s"],sender,name),
				OnAccept = function()
					self:StartReceiving(key,sender,length,name)
					self:Respond(format("RESPONSE:%s:%s",key,"YES"),sender)
				end,
				OnCancel = function() self:Respond(format("RESPONSE:%s:%s",key,"NO"),sender) end,
				button1 = L["Accept"],
				button2 = L["Reject"],
				timeout = 25,
				whileDead = 1,
				hideOnEscape = 1,
			}
			StaticPopupDialogs[popupkey] = STATIC_CONFIRM
		end
		StaticPopup_Show(popupkey)

	-- Someone responded to your send
	elseif type == "RESPONSE" and dist == "WHISPER" then
		local key,answer = split(":",args)
		local ul = Uploads[key]
		if not ul then
			return
		end
		if answer == "YES" then
			ul.accepts = ul.accepts + 1
			UploadQueue[#UploadQueue + 1] = sender
			self:ScheduleTimer("StartUpload",UL_WAIT,key,sender)

		elseif answer == "NO" then
			ul.declines = ul.declines + 1
		end
		self:RegisterComm(TRANSFER_PREFIX)
	end
end

----------------------------------
-- UPLOAD/DOWNLOAD HANDLERS
----------------------------------
function module:StartUpload(key,target)
	local ul = Uploads[key]
	local message = format("%s:%s:%s","UPLOAD",key,ul.serialData)

	self:SendCommMessage(TRANSFER_PREFIX, message, "WHISPER", target)
end

function module:StartReceiving(key,sender,length,name)

	Downloads[key] = {
		key = key,
		sender = sender,
		length = length,
		name = name,
		timer = self:ScheduleTimer("DLTimeout",DL_TIMEOUT,key,sender),
	}

	self:RegisterComm(TRANSFER_PREFIX)
end

----------------------------------
-- TRANSFERS
----------------------------------
function module:Transfer(msg, dist, sender)
	local type,args = match(msg,"(%w+):(.+)")

	-- Receiving an upload
	if type == "UPLOAD" then
		local key,serialData = split(":",args)
		local length = len(serialData)

		local dl = Downloads[key]
		if not dl or length ~= dl.length then
			return
		end

		local success, data = self:Deserialize(msg)
		-- Failed to deserialize
		if not success then
			addon:Print(format(L["Failed to load %s after downloading! Request another distribute from %s"],dl.name,dl.sender)) return
		end
		addon:UnregisterEncounter(key)
		addon:RegisterEncounter(data)

		addon.RDB[key] = data
		addon:SendWhisperComm(dl.sender,"VersionBroadcast",key,data.version)
		self:DLCompleted(key,dl.sender)


	-- Receiving COMPLETE message
	elseif type == "COMPLETE" then
		local key = args

		for i=1,#UploadQueue do
			if UploadQueue[i] == sender then
				remove(UploadQueue,i)
			end
		end

		if #UploadQueue == 0 then
			self:ULCompleted(key)
		end
	end
end


----------------------------------
-- COMPLETIONS
----------------------------------
function module:ULCompleted(key)
	addon:Print(format(L["%s upload complete"],Uploads[key].name))
end

function module:DLCompleted(key,sender)
	addon:Print(format(L["%s download from %s complete"],Downloads[key].name,sender))
	self:CancelTimer("DLTimeout",true)
	self:SendCommMessage(TRANSFER_PREFIX, format("%s:%s","COMPLETE",key), "WHISPER", sender)
	self:UnregisterComm(TRANSFER_PREFIX)
	wipe(Downloads[key])
end

----------------------------------
-- TIMEOUTS
----------------------------------
function module:ULTimeout(key)
	addon:Print(format(L["%s upload timed out"],Uploads[key].name))
end

function module:DLTimeout(key,sender)
	addon:Print(format(L["%s download from %s timed out"],Downloads[key].name,sender))
end

----------------------------------
-- COMM RECEIVED
----------------------------------

function module:OnCommReceived(prefix, msg, dist, sender)
	if sender == addon.PNAME then
		return
	end

	if prefix == MAIN_PREFIX then
		self:Main(msg, dist, sender)
	elseif prefix == TRANSFER_PREFIX then
		self:Transfer(msg, dist, sender)
	end
end

